# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from kazoo.client import KazooClient, KazooState
from kazoo.handlers.gevent import SequentialGeventHandler
from kazoo.recipe.party import Party
import kazoo.exceptions
import logging
from iris import metrics
from gevent import sleep
from itertools import cycle

logger = logging.getLogger(__name__)

UPDATE_FREQUENCY = 3


class Coordinator(object):
    def __init__(self, zk_hosts, hostname, port, join_cluster):
        self.me = '%s:%s' % (hostname, port)
        self.is_master = None
        self.slaves = cycle([])
        self.slave_count = 0
        self.started_shutdown = False

        if join_cluster:
            read_only = False
        else:
            read_only = True

        self.zk = KazooClient(hosts=zk_hosts, handler=SequentialGeventHandler(), read_only=read_only)
        event = self.zk.start_async()
        event.wait(timeout=5)

        self.lock = self.zk.Lock(path='/iris/sender_master', identifier=self.me)

        # Used to keep track of slaves / senders present in cluster
        self.party = Party(client=self.zk, path='/iris/sender_nodes', identifier=self.me)

        if join_cluster:
            self.zk.add_listener(self.event_listener)
            self.party.join()

    def am_i_master(self):
        return self.is_master

    # Used for API to get the current master
    def get_current_master(self):
        try:
            contenders = self.lock.contenders()
        except kazoo.exceptions.KazooException:
            logger.exception('Failed getting contenders')
            return None

        if contenders:
            return self.address_to_tuple(contenders[0])
        else:
            return None

    # Used for API to get the current slaves if master can't be reached
    def get_current_slaves(self):
        return [self.address_to_tuple(host) for host in self.party]

    def address_to_tuple(self, address):
        try:
            host, port = address.split(':')
            return host, int(port)
        except (IndexError, ValueError):
            logger.error('Failed getting address tuple from %s', address)
            return None

    def update_status(self):
        if self.started_shutdown:
            return

        if self.zk.state == KazooState.CONNECTED:
            if self.lock.is_acquired:
                self.is_master = True
            else:
                try:
                    self.is_master = self.lock.acquire(blocking=False, timeout=2)

                # This one is expected when we're recovering from ZK being down
                except kazoo.exceptions.CancelledError:
                    self.is_master = False

                except kazoo.exceptions.LockTimeout:
                    self.is_master = False
                    logger.exception('Failed trying to acquire lock (shouldn\'t happen as we\'re using nonblocking locks)')

                except kazoo.exceptions.KazooException:
                    self.is_master = False
                    logger.exception('ZK problem while Failed trying to acquire lock')
        else:
            logger.error('ZK connection is in %s state', self.zk.state)
            self.is_master = False

        if self.zk.state == KazooState.CONNECTED:

            if self.is_master:
                slaves = [self.address_to_tuple(host) for host in self.party if host != self.me]
                self.slave_count = len(slaves)
                self.slaves = cycle(slaves)
            else:
                self.slaves = cycle([])
                self.slave_count = 0

            # Keep us as part of the party, so the current master sees us as a slave
            if not self.party.participating:
                try:
                    self.party.join()
                except kazoo.exceptions.KazooException:
                    logger.exception('ZK problem while trying to join party')
        else:
            self.slaves = cycle([])
            self.slave_count = 0

    def update_forever(self):
        while True:
            if self.started_shutdown:
                return

            old_status = self.is_master
            self.update_status()
            new_status = self.is_master

            if old_status != new_status:
                log = logger.info
            else:
                log = logger.debug

            if self.is_master:
                log('I am the master sender')
            else:
                log('I am a slave sender')

            metrics.set('slave_instance_count', self.slave_count)
            metrics.set('is_master_sender', int(self.is_master is True))

            sleep(UPDATE_FREQUENCY)

    def leave_cluster(self):
        self.started_shutdown = True

        # cancel any attempts to acquire master lock which could make us hang
        self.lock.cancel()

        if self.zk.state == KazooState.CONNECTED:
            if self.party and self.party.participating:
                logger.info('Leaving party')
                self.party.leave()
            if self.lock and self.lock.is_acquired:
                logger.info('Releasing lock')
                self.lock.release()

        # Make us not the master
        self.is_master = False

        # Avoid sending metrics that we are still the master when we're not
        metrics.set('is_master_sender', 0)

    def event_listener(self, state):
        if state == KazooState.LOST or state == KazooState.SUSPENDED:
            logger.info('ZK state transitioned to %s. Resetting master status.', state)

            # cancel pending attempts to acquire lock which will break and leave
            # us in bad state
            self.lock.cancel()

            # make us try to re-acquire lock during next iteration when we're connected
            if self.lock.is_acquired:
                self.lock.is_acquired = False

            # make us try to rejoin the party during next iteration when we're connected
            if self.party.participating:
                self.party.participating = False

            # in the meantime we're not master
            self.is_master = None



from iris.vendors import IrisVendorManager


def test_send_through_dummy():
    vendor_manager = IrisVendorManager([{'type': 'iris_dummy'}], ['dummy_app'])
    assert vendor_manager.send_message({'mode': 'call'}) == 1
    assert vendor_manager.send_message({'application': 'dummy app', 'mode': 'call'}) == 2



class dummy_app(object):
    name = 'dummy app'

    def __init__(self, vendor):
        self.vendor = vendor

        # Contrived example of how a sample application can change how a
        # message is sent
        self.time_taken = 2

    def send(self, message):
        return self.vendor.send(message, {'time_taken': self.time_taken})



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

__all__ = []

from .core import init_plugins, find_plugin  # noqa



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import logging
import requests
import time
from iris.constants import HIPCHAT_SUPPORT

logger = logging.getLogger(__name__)


class iris_hipchat(object):
    supports = frozenset([HIPCHAT_SUPPORT])

    def __init__(self, config):
        self.config = config
        self.modes = {
            'hipchat': self.send_message
        }
        self.proxy = None
        if 'proxy' in self.config:
            host = self.config['proxy']['host']
            port = self.config['proxy']['port']
            self.proxy = {'http': 'http://%s:%s' % (host, port),
                          'https': 'https://%s:%s' % (host, port)}
        self.token = self.config.get('auth_token')
        self.room_id = int(self.config.get('room_id'))
        self.debug = self.config.get('debug')
        self.endpoint_url = self.config.get('base_url')
        self.timeout = config.get('timeout', 10)

        self.headers = {
            'Content-type': 'application/json',
        }

    def get_message_payload(self, message, mention):
        """Send notification to specified HipChat room"""
        clean_message = "{0} {1}".format(mention, message['body'])
        message_dict = {
            'message': clean_message,
            'color': 'red',
            'notify': 'true',
            'message_format': "text",
        }
        return message_dict

    def parse_destination(self, destination):
        """Determine room_id, token and user to mention
           We accept 3 formats:
             - Just a mention (@testuser)
             - Room_id and Token (12341;a20asdfgjahdASDfaskw)
             - Room_id, Token and mention (12341;a20asdfgjahdASDfaskw;@testuser)
        """
        room_id = self.room_id
        token = self.token
        mention = ""

        if destination.startswith("@"):
            mention = destination
        elif ";" in destination:
            dparts = destination.split(";")
            if len(dparts) == 3:
                try:
                    room_id = int(dparts[0])
                except ValueError:
                    logger.error("Invalid destination: %s. Using default room_id", destination)
                token = dparts[1]
                mention = dparts[2]
            elif len(dparts) == 2:
                try:
                    room_id = int(dparts[0])
                except ValueError:
                    logger.error("Invalid destination: %s. Using default room_id", destination)
                token = dparts[1]
            else:
                logger.error("Invalid destination: %s. Contains %s fields.", destination, len(dparts))
        else:
            logger.error("Invalid destination: %s. Neither @user or ; separated value.", destination)

        return room_id, token, mention

    def send_message(self, message):
        start = time.time()
        room_id, token, mention = self.parse_destination(message['destination'])

        payload = self.get_message_payload(message, mention)
        notification_url = '{0}/v2/room/{1}/notification'.format(self.endpoint_url, room_id)
        params = {'auth_token': token}

        if self.debug:
            logger.info('debug: %s', payload)
        else:
            try:
                response = requests.post(notification_url,
                                         headers=self.headers,
                                         params=params,
                                         json=payload,
                                         proxies=self.proxy,
                                         timeout=self.timeout)
                if response.status_code == 200 or response.status_code == 204:
                    return time.time() - start
                else:
                    logger.error('Failed to send message to hipchat: %d',
                                 response.status_code)
                    logger.error("Response: %s", response.content)
            except Exception as err:
                logger.exception('Hipchat post request failed: %s', err)

    def send(self, message, customizations=None):
        return self.modes[message['mode']](message)



from gevent import spawn, monkey, sleep, socket
monkey.patch_all()

import pytest  # noqa

zk_address = ('localhost', 2181)
zk_url = '%s:%s' % zk_address


class TestFailover():
    @classmethod
    def setup_class(cls):
        cls.instances = {}

    @classmethod
    def teardown_class(cls):
        for instance in cls.instances.values():
            instance.leave_cluster()

    def test_failover(self):
        from iris.coordinator.kazoo import Coordinator

        # If we can't connect to zk, skip
        try:
            sock = socket.socket()
            sock.connect(zk_address)
            sock.close()
        except socket.error:
            pytest.skip('Skipping this test as ZK server is not running/reachable.')

        # Create an initial instance which should become master
        self.instances['c1'] = Coordinator(zk_url, 'testinstance', '1001', True)
        spawn(self.instances['c1'].update_forever)
        sleep(3)
        assert self.instances['c1'].am_i_master()

        # Make another instance which should become slave
        self.instances['c2'] = Coordinator(zk_url, 'testinstance', '1002', True)
        spawn(self.instances['c2'].update_forever)
        sleep(3)
        assert self.instances['c2'].am_i_master() is False

        # Verify it became slave
        sleep(3)
        assert self.instances['c1'].slave_count == 1
        assert next(self.instances['c1'].slaves) == ('testinstance', 1002)

        # Verify API can see these instances
        self.instances['api'] = Coordinator(zk_url, None, None, False)
        assert self.instances['api'].get_current_master() == ('testinstance', 1001)
        assert ('testinstance', 1002) in self.instances['api'].get_current_slaves()

        # Kill off first master and see if slave becomes master with no slaves
        self.instances['c1'].leave_cluster()
        sleep(5)
        assert self.instances['c2'].am_i_master()
        assert self.instances['c2'].slave_count == 0

        # Start old master again and see if it becomes a slave
        self.instances['c1'] = Coordinator(zk_url, 'testinstance', '1001', True)
        spawn(self.instances['c1'].update_forever)
        sleep(5)
        assert self.instances['c1'].am_i_master() is False

        # It should show up as a slave to self.instances['c2'] which is now master
        assert self.instances['c2'].am_i_master()
        assert self.instances['c2'].slave_count == 1
        assert next(self.instances['c2'].slaves) == ('testinstance', 1001)


def test_non_cluster():
    from iris.coordinator.noncluster import Coordinator

    assert Coordinator(False, []).am_i_master() is False

    master_without_slaves = Coordinator(True, [])
    assert master_without_slaves.am_i_master()
    assert master_without_slaves.slave_count == 0

    master_with_slaves = Coordinator(True, [{'host': 'testinstance', 'port': 1001}, {'host': 'testinstance', 'port': 1002}])
    assert master_with_slaves.am_i_master()
    assert master_with_slaves.slave_count == 2

    slaves = master_with_slaves.slaves
    assert next(slaves) == ('testinstance', 1001)
    assert next(slaves) == ('testinstance', 1002)
    assert next(slaves) == ('testinstance', 1001)
    assert next(slaves) == ('testinstance', 1002)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from falcon import HTTPBadRequest


def test_parse_valid_body():
    from iris.webhooks.alertmanager import alertmanager
    am_webhook = alertmanager()

    fake_post = {
        "receiver": "iris-test",
        "status": "firing",
        "alerts": [{
            "status": "firing",
            "labels": {
                "alertname": "InstanceDown",
                "group": "testgroup",
                "instance": "testinstance:8080",
                "job": "testjob",
                "monitor": "testmonitor",
                "severity": "page"
            },
            "annotations": {
                "description": "testannotation",
                "summary": "testsummary"
            },
            "startsAt": "2017-12-04T09:17:22.804Z",
            "endsAt": "0001-01-01T00:00:00Z",
            "generatorURL": "http://1a80c99c0865:9090/graph?g0.expr=up+%3D%3D+0\\u0026g0.tab=0"
        }],
        "groupLabels": {
            "alertname": "testalert",
            "iris_plan": "testplan"
        },
        "commonLabels": {
            "alertname": "testalert",
            "group": "testgroup",
            "instance": "testinstance:8080",
            "job": "testjob",
            "monitor": "testmonitor",
            "severity": "page"
        },
        "commonAnnotations": {
            "description": "testcommonannotation",
            "summary": "testcommonsummary"
        },
        "externalURL": "http://a9fabe732a60:9093",
        "version": "4",
        "groupKey": "{}:{alertname=\"testalert\"}"
    }
    am_webhook.validate_post(fake_post)


def test_parse_invalid_body():
    from iris.webhooks.alertmanager import alertmanager
    am_webhook = alertmanager()

    fake_post = {
        "receiver": "iris-test",
        "status": "firing",
        "alerts": [{
            "status": "firing",
            "labels": {
                "alertname": "InstanceDown",
                "group": "testgroup",
                "instance": "testinstance:8080",
                "job": "testjob",
                "monitor": "testmonitor",
                "severity": "page"
            },
            "annotations": {
                "description": "testannotation",
                "summary": "testsummary"
            },
            "startsAt": "2017-12-04T09:17:22.804Z",
            "endsAt": "0001-01-01T00:00:00Z"
        }],
        "groupLabels": {
            "alertname": "testalert"
        },
        "commonLabels": {
            "alertname": "testalert"
        },
        "commonAnnotations": {
            "description": "testcommonannotation"
        },
        "groupKey": "{}:{alertname=\"testalert\"}"
    }
    try:
        am_webhook.validate_post(fake_post)
    except HTTPBadRequest:
        pass



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from . import db
import logging

logger = logging.getLogger(__name__)

applications = {}  # name -> dict of info
priorities = {}    # name -> dict of info
target_types = {}  # name -> id
target_roles = {}  # name -> id
modes = {}         # name -> id


def cache_applications():
    global applications
    connection = db.engine.raw_connection()
    cursor = connection.cursor(db.dict_cursor)
    cursor.execute('''SELECT `name`, `id`, `key`, `allow_other_app_incidents`,
                             `allow_authenticating_users`, `secondary_key`
                      FROM `application`''')
    apps = cursor.fetchall()
    new_applications = {}
    for app in apps:
        cursor.execute(
            'SELECT `name` FROM `template_variable` WHERE `application_id` = %s',
            app['id'])
        app['variables'] = [row['name'] for row in cursor]
        cursor.execute('''SELECT `mode`.`name`
                          FROM `mode`
                          JOIN `application_mode` on `mode`.`id` = `application_mode`.`mode_id`
                          WHERE `application_mode`.`application_id` = %s''', app['id'])
        app['supported_modes'] = [row['name'] for row in cursor]
        cursor.execute('''SELECT `mode`.`name` AS mode_name, `application_custom_sender_address`.`sender_address` AS address
                          FROM `application_custom_sender_address`
                          JOIN `mode` on `mode`.`id` = `application_custom_sender_address`.`mode_id`
                          WHERE `application_custom_sender_address`.`application_id` = %s''', app['id'])
        app['custom_sender_addresses'] = {row['mode_name']: row['address'] for row in cursor}
        cursor.execute('''SELECT `notification_category`.`id`, `notification_category`.`name`,
                                 `notification_category`.`description`, `notification_category`.`mode_id`,
                                 `mode`.`name` AS mode
                          FROM `notification_category`
                          JOIN `mode` ON `notification_category`.`mode_id` = `mode`.`id`
                          WHERE `application_id` = %s''', app['id'])
        app['categories'] = {row['name']: row for row in cursor}
        new_applications[app['name']] = app
    applications = new_applications
    connection.close()
    cursor.close()
    logger.debug('Loaded applications: %s', ', '.join(applications))


def cache_priorities():
    global priorities
    connection = db.engine.raw_connection()
    cursor = connection.cursor(db.dict_cursor)
    cursor.execute('''SELECT `priority`.`id`, `priority`.`name`, `priority`.`mode_id`
                      FROM `priority`''')
    priorities = {row['name']: row for row in cursor}
    cursor.close()
    connection.close()


def cache_target_types():
    global target_types
    connection = db.engine.raw_connection()
    cursor = connection.cursor()
    cursor.execute('''SELECT `name`, `id` FROM target_type''')
    target_types = dict(cursor)
    cursor.close()
    connection.close()


def cache_target_roles():
    global target_roles
    connection = db.engine.raw_connection()
    cursor = connection.cursor()
    cursor.execute('''SELECT `name`, `id` FROM target_role''')
    target_roles = dict(cursor)
    cursor.close()
    connection.close()


def cache_modes():
    global modes
    connection = db.engine.raw_connection()
    cursor = connection.cursor()
    cursor.execute('''SELECT `name`, `id` FROM mode''')
    modes = dict(cursor)
    cursor.close()
    connection.close()


def init():
    cache_applications()
    cache_priorities()
    cache_target_types()
    cache_target_roles()
    cache_modes()



import requests
import base64
import hashlib
import hmac
import time
import logging

logger = logging.getLogger(__name__)
logging.getLogger('requests').setLevel(logging.WARNING)


class IrisAuth(requests.auth.AuthBase):
    def __init__(self, app, key):
        if not isinstance(app, bytes):
            app = app.encode('utf-8')
        self.header = b'hmac ' + app + b':'
        if not isinstance(key, bytes):
            key = key.encode('utf-8')
        self.HMAC = hmac.new(key, b'', hashlib.sha512)

    def __call__(self, request):
        HMAC = self.HMAC.copy()

        path = request.path_url.encode('utf8')
        method = request.method.encode('utf8')
        body = request.body or b''
        window = str(int(time.time()) // 5).encode('utf8')
        HMAC.update(b'%s %s %s %s' % (window, method, path, body))

        digest = base64.urlsafe_b64encode(HMAC.digest())
        request.headers['Authorization'] = self.header + digest
        return request


class IrisClient(requests.Session):
    def __init__(self, base, version=0, iris_app=None, iris_app_key=None):
        super(IrisClient, self).__init__()
        self.url = base + '/v%d/' % version
        adapter = requests.adapters.HTTPAdapter(pool_connections=100, pool_maxsize=100)
        self.mount('http://', adapter)
        self.mount('https://', adapter)

        if iris_app and iris_app_key:
            self.auth = IrisAuth(iris_app, iris_app_key)
            logger.info('Initializing iris api client with auth using app %s', iris_app)
        else:
            logger.warning('Initializing iris api client without auth')

    def get(self, path, *args, **kwargs):
        return super(IrisClient, self).get(self.url + path, *args, **kwargs)

    def post(self, path, *args, **kwargs):
        return super(IrisClient, self).post(self.url + path, *args, **kwargs)

    def put(self, path, *args, **kwargs):
        return super(IrisClient, self).put(self.url + path, *args, **kwargs)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from iris.custom_import import import_custom_module
from gevent import sleep
import logging
logger = logging.getLogger(__name__)

stats_reset = {}
stats = {}

metrics_provider = None


def get_metrics_provider(config, app_name):
    return import_custom_module('iris.metrics', config['metrics'])(config, app_name)


def emit():
    if metrics_provider:
        metrics_provider.send_metrics(stats)
    stats.update(stats_reset)


# It's expected that you gevent.spawn this after you monkeypatch everything
def emit_forever():
    while True:
        emit()
        sleep(60)


def init(config, app_name, default_stats):
    global metrics_provider
    metrics_provider = get_metrics_provider(config, app_name)
    stats_reset.update(default_stats)
    stats.update(stats_reset)


def add_new_metrics(default_stats):
    stats_reset.update(default_stats)

    # avoid clobbering existing metrics if they are already present
    for key, default_value in default_stats.items():
        stats.setdefault(key, default_value)


def incr(key, inc=1):
    try:
        stats[key] += inc
    except KeyError as e:
        logger.exception('failed incrementing nonexistent metric: %s', e)


def set(key, value):
    stats[key] = value



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from iris.custom_import import import_custom_module
from collections import defaultdict
import logging
import itertools
import random
import copy
logger = logging.getLogger(__name__)


class IrisVendorException(Exception):
    pass


class IrisVendorManager():

    def __init__(self, vendors, application_vendors):
        self.max_tries_per_message = 5
        self.vendors_iter = {}
        self.app_specific_vendors_iter = defaultdict(dict)
        self.all_vendor_instances = []

        applications = {}

        for application in application_vendors:
            try:
                application_cls = import_custom_module('iris.applications', application)
            except ImportError:
                logger.exception('Failed importing application %s', application)
                continue

            try:
                applications[application_cls.name] = application_cls
            except AttributeError:
                logger.exception('Failed loading name of custom application %s', application)
                continue

        vendor_instances = defaultdict(list)
        app_vendor_instances = defaultdict(lambda: defaultdict(list))
        for vendor_config in vendors:
            vendor_cls = import_custom_module('iris.vendors', vendor_config['type'])
            for mode in vendor_cls.supports:
                vendor_instance = vendor_cls(copy.deepcopy(vendor_config))
                vendor_instances[mode].append(vendor_instance)
                for application_name, application_cls in applications.items():
                    app_vendor_instances[application_name][mode].append(application_cls(vendor_instance))

        for mode, instances in vendor_instances.items():
            random.shuffle(instances)
            self.all_vendor_instances += instances
            self.vendors_iter[mode] = itertools.cycle(instances)

        for application_name, modes in app_vendor_instances.items():
            for mode_name, instances in modes.items():
                random.shuffle(instances)
                self.app_specific_vendors_iter[application_name][mode_name] = itertools.cycle(instances)

    def send_message(self, message):
        for tries, vendor in enumerate(self.app_specific_vendors_iter.get(message.get('application'), self.vendors_iter)[message['mode']]):
            if tries > self.max_tries_per_message:
                logger.warning('Exhausted %d tries for message %s', tries, message)
                break
            logger.debug('Attempting %s send using vendor %s', message['mode'], vendor)
            try:
                return vendor.send(message)
            except Exception:
                logger.exception('Sending %s with vendor %s failed', message, vendor)
                continue

        raise IrisVendorException('All %s vendors failed for %s' % (message['mode'], message))

    def cleanup(self):
        for vendors in self.all_vendor_instances:
            cleanup_method = getattr(vendors, 'cleanup', None)
            if cleanup_method:
                cleanup_method()



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import logging
logger = logging.getLogger(__name__)


class dummy(object):
    def __init__(self, config, appname):
        self.appname = appname

    def send_metrics(self, metrics):
        logging.info('Sending metrics: %s', ', '.join('%s: %s' % (key, metrics[key]) for key in sorted(metrics)))



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import logging
import requests
import time
from iris.constants import SMS_SUPPORT, CALL_SUPPORT

logger = logging.getLogger(__name__)


class iris_messagebird(object):
    supports = frozenset([SMS_SUPPORT, CALL_SUPPORT])

    def __init__(self, config):
        self.config = config
        self.modes = {
            SMS_SUPPORT: self.send_message,
            CALL_SUPPORT: self.send_call,
        }

        self.debug = self.config.get('debug')

        self.proxy = None
        if 'proxy' in self.config:
            host = self.config['proxy']['host']
            port = self.config['proxy']['port']
            self.proxy = {'http': 'http://%s:%s' % (host, port),
                          'https': 'https://%s:%s' % (host, port)}

        base_url = 'https://rest.messagebird.com'
        self.endpoint_url_messages = base_url + '/messages'
        self.endpoint_url_voicemessages = base_url + '/voicemessages'

        self.access_key = self.config.get('access_key')
        self.headers = {
            'Accept': 'application/json',
            'Authorization': 'AccessKey ' + self.access_key,
            'User-Agent': 'Iris',
            'Content-Type': 'application/json'
        }
        self.timeout = config.get('timeout', 10)

    def get_message_payload(self, message):
        """Format a proper message dict"""
        message_dict = {
            'body': message['body'],
            'originator': 'Iris',
            'recipients': message['destination'],
        }
        if message['mode'] == 'call':
            message_dict['repeat'] = 3

        return message_dict

    def send_message(self, message):
        start = time.time()
        payload = self.get_message_payload(message)

        if self.debug:
            logger.info('debug: %s', payload)
        else:
            try:
                response = requests.post(self.endpoint_url_messages,
                                         headers=self.headers,
                                         json=payload,
                                         proxies=self.proxy,
                                         timeout=self.timeout)
                if response.status_code == 201:
                    return time.time() - start
                else:
                    logger.error('Failed to send message to messagebird: %d. Response: %s',
                                 response.status_code, response.content)
            except Exception as err:
                logger.exception('messagebird post request failed: %s', err)

    def send_call(self, message):
        start = time.time()
        payload = self.get_message_payload(message)

        if self.debug:
            logger.info('debug: %s', payload)
        else:
            try:
                response = requests.post(self.endpoint_url_voicemessages,
                                         headers=self.headers,
                                         json=payload,
                                         proxies=self.proxy,
                                         timeout=self.timeout)
                if response.status_code == 201:
                    return time.time() - start
                else:
                    logger.error('Failed to send voicemessage to messagebird: %d. Response: %s',
                                 response.status_code, response.content)
            except Exception as err:
                logger.exception('messagebird voicemessage post request failed: %s', err)

    def send(self, message, customizations=None):
        return self.modes[message['mode']](message)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import setuptools
import re

with open('src/iris/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE).group(1)

with open('README.md', 'r') as fd:
    long_description = fd.read()

setuptools.setup(
    name='irisapi',
    version=version,
    description='Iris is a highly configurable and flexible service for paging and messaging.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/linkedin/iris',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3'
    ],
    package_dir={'': 'src'},
    packages=setuptools.find_packages('src'),
    include_package_data=True,
    install_requires=[
        'streql==3.0.2',
        'dnspython',
        'phonenumbers==7.4.1',
        'twilio==6.25.0',
        'google-api-python-client==1.4.2',
        'oauth2client==1.4.12',
        'slackclient==0.16',
        'PyYAML',
        'greenlet==0.4.16',
        'gevent==1.4.0',
        'falcon==1.4.1',
        'falcon-cors',
        'ujson==1.35',
        'requests',
        'PyMySQL==0.9.3',
        'SQLAlchemy==1.3.0',
        'Jinja2',
        'Markdown',
        'click',
        'msgpack==1.0.0',
        'cssmin',
        'beaker',
        'cryptography==2.3',
        'webassets',
        'python-ldap==3.1.0',
        'exchangelib==2.2.0',
        'setproctitle',
        'pyfcm==1.4.3',
        'oncallclient==1.0.0',
        'idna==2.7',
        'pyqrcode==1.2.1'
    ],
    extras_require={
        'kazoo': ['kazoo==2.6.1'],
        # plugin deps
        'influxdb': ['influxdb'],
        'prometheus': ['prometheus_client'],
        'dev': [
            'gunicorn',
            'Sphinx==1.5.6',
            'sphinxcontrib-httpdomain',
            'sphinx_rtd_theme',
            # test deps
            'mock==2.0.0',
            'pytest==5.2.0',
            'pytest-mock==1.5.0',
            'pytest-cov',
            'flake8==3.5.0',
            'tox',
            'requests-mock==1.1.0',
        ],
    },
    entry_points={
        'console_scripts': [
            'iris-dev = iris.bin.run_server:main',
            'iris = iris.bin.run_server:main',
            'iris-sender = iris.bin.sender:main',
            'iris-owa-sync = iris.bin.owasync:main',
            'iris-sync-targets = iris.bin.sync_targets:main',
            'iris-process-retention = iris.bin.retention:main',
            'iris-app-stats = iris.bin.app_stats:main',
            'iris_ctl = iris.bin.iris_ctl:main',
            'build_assets = iris.bin.ui_build_assets:main',
        ]
    }
)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from collections import deque
import jinja2
from jinja2.sandbox import SandboxedEnvironment
from gevent import spawn, sleep
from gevent.pool import Pool
from .message import update_message_mode
from .. import db
from ..role_lookup import get_role_lookups
from . import auditlog
from ..client import IrisClient

import requests
import logging
logger = logging.getLogger(__name__)


iris_client = None
plans = None
templates = None
incidents = None
roles = None
targets = None
plan_notifications = None
target_reprioritization = None
target_names = None
targets_for_role = None
dynamic_plan_map = None


class Cache():
    def __init__(self, engine, sql, active):
        self.engine = engine
        self.sql = sql
        self.active = active
        self.data = {}

    def __getitem__(self, key):
        try:
            return self.data[key]
        except KeyError:
            connection = self.engine.raw_connection()
            cursor = connection.cursor(db.dict_cursor)
            cursor.execute(self.sql, key)
            ret = self.data[key] = cursor.fetchone()
            cursor.close()
            connection.close()
            return ret

    def purge(self):
        if self.data and self.active:
            connection = self.engine.raw_connection()
            cursor = connection.cursor()
            cursor.execute(self.active, [tuple(self.data)])
            for key in self.data.keys() - {row[0] for row in cursor}:
                del self.data[key]
            cursor.close()
            connection.close()
        else:
            self.data = {}


class DynamicPlanMap(Cache):
    def __getitem__(self, key):
        try:
            return self.data[key]
        except KeyError:
            connection = self.engine.raw_connection()
            cursor = connection.cursor(db.dict_cursor)
            cursor.execute(self.sql, key)
            ret = self.data[key] = {row['dynamic_index']: row for row in cursor}
            cursor.close()
            connection.close()
            return ret


class Templates():
    def __init__(self, engine):
        # Autoescape needs to be False to avoid html-encoding ampersands in emails. This
        # does not create security vulns as on the frontend, html is escaped using Handlebars,
        # and in emails, html is allowed and ampersands are escaped with markdown's renderer.
        self.env = SandboxedEnvironment(autoescape=False)

        self.engine = engine
        self.active = {}
        self.data = {}

    def __getitem__(self, key):
        try:
            return self.data[key]
        except KeyError:
            template = {}
            connection = self.engine.raw_connection()
            cursor = connection.cursor()
            cursor.execute('''
                SELECT `template`.`id`, `application`.`name`, `mode`.`name`,
                       `template_content`.`subject`, `template_content`.`body`
                FROM `template_active`
                JOIN `template` ON `template`.`id` = `template_active`.`template_id`
                JOIN `template_content` ON `template_content`.`template_id` = `template_active`.`template_id`
                JOIN `application` ON `template_content`.`application_id` = `application`.`id`
                JOIN `mode` ON `template_content`.`mode_id` = `mode`.`id`
                WHERE `template_active`.`name` = %s''', key)

            for template_id, application, mode, subject, body in cursor:
                logger.debug('[+] adding template: %s %s %s %s', key, template_id, application, mode)
                try:
                    # make sure message_id is delivered to the user
                    if self.has_message_id(subject) or self.has_message_id(body):
                        subject = self.env.from_string(subject)
                    else:
                        if subject:
                            subject = self.env.from_string('{{ iris.message_id }} ' + subject)
                        else:
                            subject = self.env.from_string('{{ iris.message_id }}')
                    body = self.env.from_string(body)
                except jinja2.exceptions.TemplateSyntaxError:
                    logger.info('[-] error parsing template: %s %s %s %s', key, template_id, application, mode)
                    continue
                template['id'] = template_id
                template.setdefault(application, {})[mode] = {
                    'subject': subject,
                    'body': body
                }

            self.data[key] = template
            cursor.close()
            connection.close()
            return template

    def has_message_id(self, source):
        valid = False
        ast = self.env.parse(source)
        for node in ast.body:
            try:
                nodes = node.nodes
            except AttributeError:
                continue
            for n in nodes:
                try:
                    if n.node.name == 'iris':
                        try:
                            if n.arg.value == 'message_id':
                                valid = True
                        except AttributeError:
                            if n.attr == 'message_id':
                                valid = True
                except AttributeError:
                    continue
                if not valid:
                    break
            if not valid:
                break

        return valid

    def refresh(self):
        logger.info('refreshing templates')

        try:
            request = iris_client.get('templates/?active=1&fields=id&fields=name')
            request.raise_for_status()
            templates_response = request.json()
        except (requests.exceptions.RequestException, ValueError):
            logger.exception('Failed to hit api to get active templates')
            return

        active = {item['id']: item['name'] for item in templates_response}

        new_active_ids = active.keys()
        old_active_ids = self.active.keys()

        old_ids = old_active_ids - new_active_ids
        new_ids = new_active_ids - old_active_ids

        for template_id in old_ids:
            try:
                del self.data[self.active[template_id]]
            except KeyError:
                logger.exception('Failed pruning old template_id %s', template_id)

        Pool(30).map(self.__getitem__, (active[id] for id in new_ids))

        self.active = active


class Plans():
    def __init__(self, engine):
        self.engine = engine
        self.active = {}
        self.data = {}
        # Autoescape turned off since HTML characters can exist in plan tracking template.
        # See details in Templates()
        self.template_env = SandboxedEnvironment(autoescape=False)

    def __getitem__(self, key):
        try:
            return self.data[key]
        except KeyError:
            fields = ['threshold_window', 'threshold_count', 'aggregation_window', 'creator',
                      'aggregation_reset', 'name', 'tracking_type', 'tracking_key', 'tracking_template']

            try:
                request = iris_client.get('plans/%s/' % key, params={'fields': fields})
                request.raise_for_status()
                plan = request.json()
            except (requests.exceptions.RequestException, ValueError):
                logger.exception('Failed to hit api to get plan %s', key)
                return None

            logger.debug('[+] adding plan: %s', key)

            steps = {}
            for idx, notifications in enumerate(plan['steps']):
                steps[idx + 1] = [n['id'] for n in notifications]
            plan['steps'] = steps

            if plan['tracking_template']:
                tracking_template = plan['tracking_template']
                if plan['tracking_type'] == 'email':
                    for application, application_templates in tracking_template.items():
                        try:
                            tracking_template[application] = {
                                'email_subject': self.template_env.from_string(application_templates['email_subject']),
                                'email_text': self.template_env.from_string(application_templates['email_text']),
                            }
                            html_template = application_templates.get('email_html')
                            if html_template:
                                tracking_template[application]['email_html'] = self.template_env.from_string(html_template)
                        except jinja2.exceptions.TemplateSyntaxError:
                            logger.exception('[-] error parsing Plan template for %s: %s', key, application)
                            continue
                else:
                    for application, application_templates in tracking_template.items():
                        try:
                            tracking_template[application] = {
                                'body': self.template_env.from_string(application_templates['body']),
                            }
                        except jinja2.exceptions.TemplateSyntaxError:
                            logger.exception('[-] error parsing Plan template for %s: %s', key, application)
                            continue
                plan['tracking_template'] = tracking_template

            self.data[key] = plan
            return plan

    def refresh(self):
        logger.info('refreshing plans')

        try:
            request = iris_client.get('plans/?active=1&fields=id&fields=name')
            request.raise_for_status()
            plans_response = request.json()
        except (requests.exceptions.RequestException, ValueError):
            logger.exception('Failed to hit api to refresh plans')
            return

        active = {item['id']: item['name'] for item in plans_response}

        new_active_ids = active.keys()
        old_active_ids = self.active.keys()

        old_ids = old_active_ids - new_active_ids
        new_ids = new_active_ids - old_active_ids

        for plan_id in old_ids:
            try:
                del self.data[plan_id]
            except KeyError:
                logger.exception('Failed pruning old plan_id %s', plan_id)

        Pool(30).map(self.__getitem__, (active[id] for id in new_ids))

        self.active = active


class TargetReprioritization(object):
    def __init__(self, engine):
        self.engine = engine
        # (target, src_mode): (dst_mode, destination, count, deque)
        self.rates = {}

    def refresh(self):
        while True:
            rates = {}
            connection = db.engine.raw_connection()
            cursor = connection.cursor()
            cursor.execute(
                '''SELECT
                    `target`.`name`,
                    `src_mode`.`name`,
                    `dst_mode`.`name`,
                    `target_contact`.`destination`,
                    `target_reprioritization`.`count`,
                    `target_reprioritization`.`duration`
                FROM `target_reprioritization`
                JOIN `target` ON `target_reprioritization`.`target_id` = `target`.`id`
                JOIN `mode` as `src_mode` ON `target_reprioritization`.`src_mode_id` = `src_mode`.`id`
                JOIN `mode` as `dst_mode` ON `target_reprioritization`.`dst_mode_id` = `dst_mode`.`id`
                JOIN `target_contact` ON `target_contact`.`target_id`=`target`.`id` AND `target_contact`.`mode_id` = `dst_mode`.`id`'''
            )
            for target, src_mode, dst_mode, destination, count, duration in cursor:
                if destination is None:
                    logger.info('invalid target reprioritization rule (%s, %s): (%s, %s, %d, %d)',
                                target, src_mode, dst_mode, destination, count, duration)
                    continue
                rates[(target, src_mode)] = (dst_mode, destination, count, duration / 60)

            cursor.close()
            connection.close()

            current = rates.keys()
            old = self.rates.keys()

            # purge old rate entries
            for key in old - current:
                logger.debug('deleting target reprioritization rule for %r: %r', key, self.rates[key])
                del self.rates[key]

            # new rate entries:
            for key in current - old:
                logger.debug('creating target reprioritization rule for %r: %r', key, rates[key])
                dst_mode, destination, count, duration_min = rates[key]
                self.rates[key] = (dst_mode, destination, count, deque([0] * duration_min, maxlen=duration_min))

            # loop through existing entries
            # if the settings have changed, gracefully alter them
            # either way, push in a new bucket
            for key in current & old:
                current_dst_mode, current_destination, current_count, current_duration_min = rates[key]
                old_dst_mode, old_destination, old_count, old_buckets = self.rates[key]
                if old_dst_mode != current_dst_mode or old_destination != current_destination or old_count != current_count or len(old_buckets) != current_duration_min:  # noqa FIXME: refactor this line
                    logger.debug('updating target reprioritization rule for %r: %r | %r', key, self.rates[key], rates[key])
                    # create a new bucket truncating if it's smaller and growing if it is larger, while preserving the counts
                    current_buckets = deque(old_buckets, maxlen=current_duration_min)
                    self.rates[key] = (current_dst_mode, current_destination, current_count, current_buckets)
                # add the new interval bucket to each
                self.rates[key][-1].append(0)

            logger.info('refreshed target reprioritization rules: %d', len(self.rates))
            logger.debug(self.rates)
            sleep(60)

    def __call__(self, message, seen=None):
        original_mode = message.get('mode')

        if not original_mode:
            return

        if seen is None:
            seen = set([original_mode])
        else:
            if original_mode in seen:
                logger.info('target reprioritization loop detected for %s, %s: %r',
                            message['target'], original_mode, seen)
                return
            else:
                seen.add(original_mode)
        try:
            dst_mode, destination, count, buckets = self.rates[(message['target'], original_mode)]
            logger.debug('reprioritization (%s, %s): (%s, %s, %d, %r)',
                         message['target'], original_mode, dst_mode, destination, count, buckets)
            # increment the bucket for this minute
            buckets[-1] += 1
            if sum(buckets) > count:
                logger.debug('target reprioritization rule triggered (%s, %s): (%s, %s, %d, %r)',
                             message['target'], original_mode, dst_mode, destination, count, buckets)
                # sum of all counts for duration exceeds count
                # reprioritize to destination mode
                message['mode'] = dst_mode
                message['destination'] = destination
                update_message_mode(message)
                self.__call__(message, seen)
                auditlog.message_change(message['message_id'], auditlog.MODE_CHANGE, original_mode, dst_mode, 'Changing mode due to reprioritization')
            else:
                return
        except KeyError:
            # target has no reprioritization rules defined
            # leave existing mode
            return


class RoleTargets():
    def __init__(self, role_lookups, engine):
        self.data = {}
        self.role_lookups = role_lookups
        self.engine = engine
        self.active_targets = set()
        self.initialize_active_targets()

    def __call__(self, role, target):
        try:
            return self.data[(role, target)]
        except KeyError:

            names = None

            # Iterate through our role lookup modules until we find one that works.
            for role_lookup in self.role_lookups:
                names = role_lookup.get(role, target)
                if names is not None:
                    break

            if names is None:
                logger.info('All role lookups modules failed to lookup %s:%s', role, target)
                self.data[(role, target)] = None
                return None

            names = self.prune_inactive_targets(names)
            self.data[(role, target)] = names
            return names

    def purge(self):
        self.data = {}
        self.initialize_active_targets()

    def prune_inactive_targets(self, usernames):
        if not usernames:
            return []
        return self.active_targets & set(usernames)

    def initialize_active_targets(self):
        connection = self.engine.raw_connection()
        cursor = connection.cursor()
        cursor.execute('SELECT `name` FROM `target` WHERE `active` = TRUE AND `target`.`type_id` = (SELECT `id` FROM `target_type` WHERE `name` = "user")')
        self.active_targets = {row[0] for row in cursor}
        cursor.close()
        connection.close()


def refresh():
    plans.refresh()
    templates.refresh()


def purge():
    targets.purge()
    target_names.purge()
    targets_for_role.purge()
    incidents.purge()
    dynamic_plan_map.purge()
    plan_notifications.purge()


def init(api_host, config):
    global targets_for_role, target_names, target_reprioritization, plan_notifications, targets
    global roles, incidents, templates, plans, iris_client, dynamic_plan_map

    iris_client = IrisClient(api_host, 0)

    # make sure API is online
    max_trey = 36
    api_chk_cnt = 0
    while api_chk_cnt < max_trey:
        try:
            re = iris_client.get('target_roles')
            if re.status_code == 200:
                break
        except Exception:
            pass
        api_chk_cnt += 1
        logger.warning(
            'Not able to connect to Iris API %s, retry in 5 seconds.',
            iris_client.url)
        sleep(5)

    if api_chk_cnt >= max_trey:
        import sys
        sys.exit('FATAL: Not able to connect to Iris API: %s' % iris_client.url)

    plans = Plans(db.engine)
    templates = Templates(db.engine)
    incidents = Cache(db.engine,
                      'SELECT * FROM `incident` WHERE `id`=%s',
                      'SELECT `id` from `incident` WHERE `active`=True AND `id` IN %s')
    roles = Cache(db.engine, 'SELECT * FROM `target_role` WHERE `id`=%s', None)
    targets = Cache(db.engine, 'SELECT * FROM `target` WHERE `id`=%s', None)
    plan_notifications = Cache(db.engine,
                               'SELECT * FROM `plan_notification` WHERE `id`=%s',
                               ('SELECT `plan_notification`.`id` FROM `plan_notification` '
                                'JOIN `plan_active` ON `plan_notification`.`plan_id` = `plan_active`.`plan_id` '
                                'AND `plan_notification`.`id` IN %s'))
    target_reprioritization = TargetReprioritization(db.engine)
    target_names = Cache(db.engine, 'SELECT * FROM `target` WHERE `name`=%s AND `active` = TRUE', None)
    role_lookups = get_role_lookups(config)
    targets_for_role = RoleTargets(role_lookups, db.engine)
    dynamic_plan_map = DynamicPlanMap(db.engine,
                                      'SELECT * FROM `dynamic_plan_map` WHERE `incident_id` = %s',
                                      '''SELECT dynamic_plan_map.* FROM `dynamic_plan_map`
                                         JOIN `incident` ON `incident`.`id` = `dynamic_plan_map`.`incident_id`
                                         WHERE `incident`.`active` = TRUE AND `incident_id` IN %s''')

    spawn(target_reprioritization.refresh)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# -*- coding:utf-8 -*-


from phonenumbers import (format_number as pn_format_number, parse as pn_parse,
                          PhoneNumberFormat)
from gevent import sleep
import datetime
import ujson
from . import db
import re
import msgpack
import logging

logger = logging.getLogger(__name__)

uuid4hex = re.compile('[0-9a-f]{32}\Z', re.I)
allowed_text_response_actions = frozenset(['suppress', 'claim'])


def normalize_phone_number(num):
    return pn_format_number(pn_parse(num, 'US'),
                            PhoneNumberFormat.INTERNATIONAL)


def validate_msg_id(msg_id):
    return msg_id.isdigit() or uuid4hex.match(msg_id)


def parse_response(response, mode, source):
    claim_all = False
    claim_last = False
    if response.lower().startswith('f'):
        return None, 'Sincerest apologies'
    # One-letter shortcuts for claim all/last
    elif re.match('^a\s*$', response, re.IGNORECASE):
        claim_all = True
    elif re.match('^l\s*$', response, re.IGNORECASE):
        claim_last = True

    # Skip message splitting for single-letter responses
    if not (claim_all or claim_last):
        halves = response.split(None, 1)

        # $id cmd (args..)
        cmd = halves[1].split()[0].lower()
        msg_id = halves[0]
        if validate_msg_id(msg_id) and (cmd in allowed_text_response_actions):
            return msg_id, cmd

        # cmd $id (args..)
        cmd = halves[0].lower()
        msg_id = halves[1].split(None, 1)
        if len(msg_id) == 1:
            args = []
        else:
            args = msg_id[1:]
        msg_id = msg_id[0]
        if validate_msg_id(msg_id) and (cmd in allowed_text_response_actions):
            return msg_id, ' '.join([cmd] + args)

    if claim_last or re.match('claim\s+last', response, re.IGNORECASE):
        target_name = lookup_username_from_contact(mode, source)
        if not target_name:
            logger.error('Failed resolving %s:%s to target name', mode, source)
            return None, 'claim'

        connection = db.engine.raw_connection()
        cursor = connection.cursor()
        cursor.execute('''SELECT `message`.`id` from `message`
                          JOIN `target` on `target`.`id` = `message`.`target_id`
                          JOIN `target_type` on `target`.`type_id` = `target_type`.`id`
                          WHERE `target`.`name` = %s
                          AND `target_type`.`name` = 'user'
                          ORDER BY `message`.`id` DESC
                          LIMIT 1''', (target_name, ))
        ret = cursor.fetchone()
        cursor.close()
        connection.close()

        msg_id = ret[0] if ret else None
        return msg_id, 'claim'

    elif claim_all or re.match('claim\s+all', response, re.IGNORECASE):
        target_name = lookup_username_from_contact(mode, source)
        if not target_name:
            logger.error('Failed resolving %s:%s to target name', mode, source)
            return None, 'claim_all'

        connection = db.engine.raw_connection()
        cursor = connection.cursor()
        cursor.execute('''SELECT `message`.`id` from `message`
                          JOIN `target` on `target`.`id` = `message`.`target_id`
                          JOIN `incident` on `incident`.`id` = `message`.`incident_id`
                          JOIN `target_type` on `target`.`type_id` = `target_type`.`id`
                          WHERE `target`.`name` = %s
                          AND `target_type`.`name` = 'user'
                          AND `incident`.`active` = TRUE''', (target_name, ))
        msg_ids = [row[0] for row in cursor]
        cursor.close()
        connection.close()

        return msg_ids, 'claim_all'

    return halves


def parse_email_response(first_line, subject, source):
    if subject and first_line in allowed_text_response_actions:
        subject_parts = subject.strip().split(None, 2)
        if len(subject_parts) > 1 and subject_parts[0] == 'Re:' and validate_msg_id(subject_parts[1]):
            return subject_parts[1], first_line
    return parse_response(first_line, 'email', source)


def get_incident_id_from_message_id(msg_id):
    sql = 'SELECT `message`.`incident_id` FROM `message` WHERE `message`.`id` = %s'
    connection = db.engine.raw_connection()
    cursor = connection.cursor()
    cursor.execute(sql, (msg_id, ))
    ret = cursor.fetchone()
    cursor.close()
    connection.close()
    return ret[0] if ret else None


def get_incident_ids_from_message_ids(msg_ids):
    sql = 'SELECT DISTINCT `message`.`incident_id` FROM `message` WHERE `message`.`id` in %s'
    connection = db.engine.raw_connection()
    cursor = connection.cursor()
    cursor.execute(sql, (tuple(msg_ids), ))
    ret = [row[0] for row in cursor]
    cursor.close()
    connection.close()
    return ret


def get_incident_context_from_message_id(msg_id):
    sql = '''SELECT `incident`.`context`
             FROM `message`
             JOIN `incident` ON `incident`.`id` = `message`.`incident_id`
             WHERE `message`.`id` = %s'''
    connection = db.engine.raw_connection()
    cursor = connection.cursor()
    cursor.execute(sql, (msg_id, ))
    ret = cursor.fetchone()
    cursor.close()
    connection.close()
    if not ret:
        return None
    return ujson.loads(ret[0])


def get_incident_context_from_batch_id(batch_id):
    """
    Return a list of contexts for a batch of messages identified by the batch id
    """
    connection = db.engine.raw_connection()
    cursor = connection.cursor()
    sql = '''SELECT `incident`.`context`
             FROM `message`
             JOIN `incident` ON `incident`.`id` = `message`.`incident_id`
             WHERE `message`.`batch` = %s'''
    cursor.execute(sql, (batch_id, ))
    contexts = [ujson.loads(row[0]) for row in cursor]
    cursor.close()
    connection.close()
    return contexts


def lookup_username_from_contact(mode, destination):
    if mode == 'sms' or mode == 'call':
        dest = normalize_phone_number(destination)
    else:
        dest = destination
    sql = '''SELECT `target`.`name` FROM `target`
             JOIN `target_contact` on `target_contact`.`target_id` = `target`.`id`
             JOIN `mode` on `mode`.`id` = `target_contact`.`mode_id`
             WHERE `mode`.`name` = %(mode)s AND `target_contact`.`destination` = %(dest)s'''

    connection = db.engine.raw_connection()
    cursor = connection.cursor()
    cursor.execute(sql, {'mode': mode, 'dest': dest})
    ret = cursor.fetchone()
    cursor.close()
    connection.close()

    return ret[0] if ret else None


def claim_incident(incident_id, owner):
    connection = db.engine.raw_connection()

    cursor = connection.cursor()
    cursor.execute('''
      SELECT `target`.`name`
      FROM `incident`
      LEFT JOIN `target` ON `target`.`id` = `incident`.`owner_id`
      WHERE `incident`.`id` = %s
    ''', (incident_id, ))
    result = cursor.fetchone()
    previous_owner = result[0] if result else None
    cursor.close()

    active = 0 if owner else 1
    now = datetime.datetime.utcnow()

    max_retries = 3

    for i in range(max_retries):
        cursor = connection.cursor()
        try:
            cursor.execute('''UPDATE `incident`
                               SET `incident`.`updated` = %(updated)s,
                                   `incident`.`active` = %(active)s,
                                   `incident`.`owner_id` = (SELECT `target`.`id` FROM `target` WHERE `target`.`name` = %(owner)s AND `type_id` = (SELECT `id` FROM `target_type` WHERE `name` = 'user'))
                               WHERE `incident`.`id` = %(incident_id)s''',
                           {'incident_id': incident_id, 'active': active, 'owner': owner, 'updated': now})

            connection.commit()
            break
        except Exception:
            logger.exception('Failed running claim query. (Try %s/%s)', i + 1, max_retries)
        finally:
            cursor.close()

        sleep(.2)

    cursor = connection.cursor()

    try:
        cursor.execute('UPDATE `message` SET `active` = 0 WHERE `incident_id`= %s', (incident_id, ))
        connection.commit()
    except Exception:
        logger.exception('Failed running query to set message for incident %s inactive', incident_id)
    finally:
        cursor.close()

    connection.close()

    return active == 1, previous_owner


def claim_bulk_incidents(incident_ids, owner):
    incident_ids = tuple(incident_ids)
    active = 0 if owner else 1
    now = datetime.datetime.utcnow()

    connection = db.engine.raw_connection()
    cursor = connection.cursor()

    cursor.execute('''UPDATE `incident`
                      SET `incident`.`updated` = %(updated)s,
                          `incident`.`active` = %(active)s,
                          `incident`.`owner_id` = (SELECT `target`.`id` FROM `target` WHERE `target`.`name` = %(owner)s AND `type_id` = (SELECT `id` FROM `target_type` WHERE `name` = 'user'))
                       WHERE `incident`.`id` IN %(incident_ids)s''',
                   {'incident_ids': incident_ids, 'active': active, 'owner': owner, 'updated': now})

    connection.commit()

    cursor.execute('UPDATE `message` SET `active` = 0 WHERE `incident_id` IN %s', (incident_ids, ))

    connection.commit()

    claimed = set()
    not_claimed = set()

    cursor.execute('SELECT `id`, `active` FROM `incident` WHERE `id` IN %s', (incident_ids, ))

    for incident_id, active in cursor:
        if active == 1:
            not_claimed.add(incident_id)
        else:
            claimed.add(incident_id)

    cursor.close()
    connection.close()

    return claimed, not_claimed


def claim_incidents_from_batch_id(batch_id, owner):
    sql = '''UPDATE `incident`
             JOIN `message` ON `message`.`incident_id` = `incident`.`id`
             SET `incident`.`owner_id` = (SELECT `target`.`id` FROM `target` WHERE `target`.`name` = %(owner)s AND `target`.`type_id` = (SELECT `id` FROM `target_type` WHERE `name` = 'user')),
                 `incident`.`updated` = %(now)s, `incident`.`active` = 0
             WHERE `message`.`batch` = %(batch_id)s'''

    args = {
        'batch_id': batch_id,
        'owner': owner,
        'now': datetime.datetime.utcnow(),
    }

    connection = db.engine.raw_connection()
    cursor = connection.cursor()

    cursor.execute(sql, args)

    connection.commit()

    cursor.execute('''UPDATE `message` `a`
                      JOIN `message` `b` ON `a`.`incident_id` = `b`.`incident_id`
                      SET `a`.`active` = 0 WHERE `b`.`batch` = %s''',
                   (batch_id, ))

    connection.commit()

    cursor.close()
    connection.close()


def msgpack_unpack_msg_from_socket(socket):
    unpacker = msgpack.Unpacker()
    while True:
        buf = socket.recv(1024)
        if not buf:
            break
        unpacker.feed(buf)
        try:
            item = next(unpacker)
        except StopIteration:
            pass
        else:
            return item


def sanitize_unicode_dict(d):
    '''Properly decode unicode strings in d to avoid breaking jinja2 renderer'''
    for key, value in d.items():
        if isinstance(value, bytes):
            try:
                d[key] = value.decode('utf-8')
            except UnicodeError:
                pass
        elif isinstance(value, dict):
            d[key] = sanitize_unicode_dict(value)
    return d



# -*- coding:utf-8 -*-
# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import os
import logging
from importlib import import_module
import yaml

logger = logging.getLogger(__name__)


def load_config_file(path=None):
    ''' Get config from path to file, defaulting to cli arg. This can easily be monkey patched. '''
    if not path:
        if len(sys.argv) <= 1:
            print('ERROR: missing config file.')
            print('usage: %s API_CONFIG_FILE' % sys.argv[0])
            sys.exit(1)
        path = sys.argv[1]

    with open(path) as h:
        return yaml.safe_load(h)


def process_config_hook(config):
    ''' Examine config dict for hooks and run them if present '''
    if 'init_config_hook' in config:
        try:
            module = config['init_config_hook']
            logger.info('Bootstrapping config using %s', module)
            getattr(import_module(module), module.split('.')[-1])(config)
        except ImportError:
            logger.exception('Failed loading config hook %s', module)

    return config


def load_config(path=None):
    '''
      Generate configs for iris in the following steps:

        * reads config from yaml file by calling `load_config_file()`.
        * reads more config values from environment variables and use them to
          override values from the config file.
        * pass config through process_config_hook(). It looks for a key called
          init_config_hook in the config, which is the name of a module to call
          which can tweak the config further.

      load_config_file() can be monkey patched, in which case this config
      loading functionality can be customized further.
    '''
    config = load_config_file(path)
    if not config:
        sys.exit('Failed to load config from ' + path)

    if 'IRIS_CFG_DB_HOST' in os.environ:
        config['db']['conn']['kwargs']['host'] = os.environ['IRIS_CFG_DB_HOST']
    if 'IRIS_CFG_DB_USER' in os.environ:
        config['db']['conn']['kwargs']['user'] = os.environ['IRIS_CFG_DB_USER']
    if 'IRIS_CFG_DB_PASSWORD' in os.environ:
        config['db']['conn']['kwargs']['password'] = os.environ['IRIS_CFG_DB_PASSWORD']
    return process_config_hook(config)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# Headers
XFRAME = ('X-Frame-Options', 'SAMEORIGIN')
XCONTENTTYPEOPTIONS = ('X-Content-Type-Options', 'nosniff')
XXSSPROTECTION = ('X-XSS-Protection', '1; mode=block')

# Avoid using magic strings when referring to the transports vendors support
SMS_SUPPORT = 'sms'
CALL_SUPPORT = 'call'
EMAIL_SUPPORT = 'email'
IM_SUPPORT = 'im'
SLACK_SUPPORT = 'slack'
HIPCHAT_SUPPORT = 'hipchat'



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.


class Authenticator:
    def __init__(self, config):
        pass

    def authenticate(self, username, password):
        return True



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from iris import db


def login_user(req, username):
    session = req.env['beaker.session']
    session['user'] = username
    session.save()


def logout_user(req):
    session = req.env['beaker.session']
    session.pop('user', None)
    session.delete()


def valid_username(username):
    connection = db.engine.raw_connection()
    cursor = connection.cursor()
    cursor.execute('''
    SELECT EXISTS(SELECT 1
                  FROM `target`
                  JOIN `user` on `target`.`id` = `user`.`target_id`
                  WHERE `name` = %s AND `type_id` = (SELECT `id` FROM `target_type` WHERE `name` = "user"))''', username)
    result = cursor.fetchone()
    cursor.close()
    connection.close()
    return result[0] == 1
